import click
from .api_wrapper import request_data, request_no_response


def create_user(username, password, permission, name, email):

    payload = {"username": "{}".format(str(username)), "password": str(password), "permissions": permission, "name": name, "email": "{}".format(str(email))}

    data = request_no_response("POST", "/users", payload=payload)

    return data


def enable_disable_user(user_id, answer):

    if answer == "enable":
        payload = {"enabled": True}
    else:
        payload = {"enabled": False}

    request_no_response("PUT", "/users/" + str(user_id) + "/enabled", payload=payload)


def change_auth_settings(user_id, payload):
    request_no_response("PUT", "/users/{}/authorizations".format(user_id), payload=payload)


def get_user_id(username):
    data = request_data("GET", "/users")
    user_id = 0
    user_uuid = 0
    for u in data["users"]:
        if username == u["username"]:
            user_id = u["id"]
            user_uuid = u['uuid']
    return user_id, user_uuid


@click.group(help="Enable, Disable or add users")
def user():
    pass


@user.command(help="Add a User to Tenable.io - User will be enabled if already exists")
@click.option("--username", "--u", default='', required=True, help="Username")
@click.option("--password", "--p", default='', required=True, help="Users password")
@click.option("--permission", "--m", default='', required=True, help="Users Permission: (16,24,32,40,64)")
@click.option("--name", "--n", default='', required=True, help="Users Name")
@click.option("--email", "--e", default='', required=True, help="Users email")
def add(username, password, permission, name, email):

    if len(password) < 11:
        click.echo("Password must be 12 chars, 1 Upper, 1 lower and 1 Special Char")
        exit()

    # Check to see has already been created
    user_id, user_uuid = get_user_id(username)

    if user_id == 0:
        # if the user doesn't exist. Create it.
        create_user(username, password, permission, name, email)
    else:
        # If the user is trying to create the user, and it already exists, Try to enable it instead
        enable_disable_user(user_id, "enable")


@user.command(help="Enable Auth settings or an Account")
@click.argument('uid')
@click.option('-account', is_flag=True, help="Enable User Account")
@click.option('-api', is_flag=True, help="Enable API Access")
@click.option('-pwd', is_flag=True, help="Enable Password Access")
@click.option('-saml', is_flag=True, help="Enable SAML Access")
def enable(uid, api, pwd, saml, account):
    payload = {"api_permitted": False, "password_permitted": False, "saml_permitted": False}
    if api:
        payload["api_permitted"] = True

    if pwd:
        payload["password_permitted"] = True

    if saml:
        payload["saml_permitted"] = True

    if api or pwd or saml:
        change_auth_settings(uid, payload)

    if account:
        enable_disable_user(uid, "enable")

    if not account and not api and not pwd and not saml:
        click.echo("\nYou need to specify an option\n\nIf you want to disable an account use the '-account' option\n")


@user.command(help="Disable a user by ID or auth settings")
@click.argument('uid')
@click.option('-account', is_flag=True, help="Enable User Account")
@click.option('-api', is_flag=True, help="Enable API Access")
@click.option('-pwd', is_flag=True, help="Enable Password Access")
@click.option('-saml', is_flag=True, help="Enable SAML Access")
def disable(uid, account, api, pwd, saml):

    payload = {"api_permitted": True, "password_permitted": True, "saml_permitted": True}

    if api:
        payload["api_permitted"] = False

    if pwd:
        payload["password_permitted"] = False

    if saml:
        payload["saml_permitted"] = False

    if api or pwd or saml:
        change_auth_settings(uid, payload)

    if account:
        enable_disable_user(uid, "disable")

    if not account and not api and not pwd and not saml:
        click.echo("\nYou need to specify an option\n\nIf you want to disable an account use the '-account' option\n")



import click
from .database import db_query
from .api_wrapper import request_data, tenb_connection
import csv
import datetime
import time
# Grab all 19506 Data

tio = tenb_connection()


def grab_hop_count(uuid):
    # grab the output of 10287 - Trace Route
    hop_count_data = db_query("select output from vulns where asset_uuid='{}' and plugin_id='10287';".format(uuid))

    # Send the raw data back
    return hop_count_data


def parse_19506_from_file(filename, scanid, histid):
    from csv import DictReader
    total_assets_scanned_list = []
    reported_scan_duration = None
    total_reported_scan_duration = None
    reported_scan_end = None
    reported_scan_start = None
    # Let's get the scatime with scan processing
    scan_history = request_data("GET", "/scans/{}/history".format(scanid))

    for hist in scan_history['history']:
        if not hist['is_archived']:
            if str(hist['id']) == str(histid):
                reported_scan_start = hist['time_start']
                reported_scan_end = hist['time_end']
                total_reported_scan_duration = reported_scan_end - reported_scan_start
                # Simple Delta from Tenable.io reported Scan times
                reported_scan_duration = str(datetime.timedelta(seconds=total_reported_scan_duration))

    # Open the file and parse the 19506 plugin
    with open(filename) as fobj:
        scan_name = "No Scan"
        scan_policy = "No Policy"
        scanner_ip = "No Scanner"
        scanner_list = []
        max_hosts = " "
        max_checks = ""
        asset_average_scantime = 0
        start_scan_timestamp_list = []
        timestamp_plus_duration_list = []

        for row in DictReader(fobj):
            plugin_output = row['Plugin Output']
            asset_uuid = row['Asset UUID']

            # split the output by return
            parsed_output = plugin_output.split("\n")

            # grab the length so we can grab the seconds
            plugin_length = len(parsed_output)

            # grab the scan duration- second to the last variable
            duration = parsed_output[plugin_length - 2]

            # Split at the colon to grab the numerical value
            intial_seconds = duration.split(" : ")

            # split to remove "secs"
            number = intial_seconds[1].split(" ")

            # grab the number for our minute calculation
            final_number = number[0]

            # For an unknown reason, the scanner will print unknown for some assets leaving no way to calulate the time.
            if final_number != 'unknown':
                try:
                    # Numerical value in seconds parsed from the plugin
                    seconds = int(final_number)

                    # Grab data pair and split it at the colon and grab the values
                    scan_name = parsed_output[9].split(" : ")[1]
                    scan_policy = parsed_output[10].split(" : ")[1]
                    scanner_ip = parsed_output[11].split(" : ")[1]
                    if scanner_ip not in scanner_list:
                        scanner_list.append(scanner_ip)
                    max_hosts = parsed_output[plugin_length - 8].split(" : ")[1]
                    max_checks = parsed_output[plugin_length - 7].split(" : ")[1]
                    # Grabbing the start time from the plugin
                    scan_time = parsed_output[plugin_length - 3].split(" : ")[1]

                    # Some timezone abrivations are not parseable with strptime. removing the timezone for those we can't
                    # parse
                    try:
                        # Set the pattern to convert into epoch
                        pattern = '%Y/%m/%d %H:%M %Z'
                        # Convert to Epoch
                        epoch = int(time.mktime(time.strptime(scan_time, pattern)))
                    except ValueError:
                        pattern = '%Y/%m/%d %H:%M'
                        # Split the time to remove the timezone 2-6 chars
                        time_less_timezone = scan_time.split(" ")
                        # merge the data and time for calculation
                        new_time = "{} {}".format(time_less_timezone[0], time_less_timezone[1])
                        # Convert to Epoch
                        epoch = int(time.mktime(time.strptime(new_time, pattern)))

                    # Add to list to calculate the first asset scanned
                    start_scan_timestamp_list.append(epoch)
                    # Add epoch and seconds to get end time.  This will be used to find the last asset scanned
                    timestamp_plus_duration_list.append(epoch + seconds)
                    # All assets and 19506 seconds in a tuple
                    total_assets_scanned_list.append((asset_uuid, seconds))
                except IndexError:
                    # This error occurs when an old scanner is used.
                    # the 19506 plugin filled with an error indicating the need for an upgrade
                    pass

        # calculate the total for AVG calc
        total = 0
        for secs in total_assets_scanned_list:
            total += secs[1]

        try:
            # Avg scantime per assset
            asset_average_scantime = total / len(total_assets_scanned_list)

            # total Scan time is: the oldest (scan_date in 19506 + scan duration in 19506) minus the earliest scan date in 19506
            total_calculated_scan_duration = (max(timestamp_plus_duration_list) - min(start_scan_timestamp_list))

            # Processing time therefore is the reported duration - the total scan duration
            processing = total_reported_scan_duration - total_calculated_scan_duration

            longest_scanned = max(total_assets_scanned_list, key=lambda uuid: uuid[1])
            shortest_scanned = min(total_assets_scanned_list, key=lambda uuid: uuid[1])

            click.echo("*" * 100)
            click.echo("\nThis data is derived from the 19506 plugins found in the {}"
                       "\nThe start time stamp(s) and the duration(s) were used to calculate the total scantime(s)\n".format(filename))
            click.echo("*" * 100)

            click.echo("\nScan Name: {}".format(scan_name))
            click.echo("Scanner Policy: {}".format(scan_policy))

            # We can't count cloud scanners; so let's not try
            if "tenable.io Scanner" in scanner_list:
                click.echo("\nScanned Using Tenable Cloud Scanners".format(len(scanner_list)))
            else:
                click.echo("\nTotal Scanners Used: {}".format(len(scanner_list)))

                for scanner in scanner_list:
                    click.echo("Scanner IP: {}".format(scanner))

            click.echo("\nMax Hosts: {}".format(max_hosts))
            click.echo("Max Checks: {}\n".format(max_checks))
            click.echo("{} {}\n".format("Total Assets scanned: ", len(total_assets_scanned_list)))

            click.echo("{:40} {:10} {}".format("Scan Stats", "H: M: S", "Asset"))
            click.echo("-" * 60)
            click.echo("{:40} {:10}".format("Total Reported Scan Duration:", reported_scan_duration))
            click.echo("{:40} {:10}".format("T.io Processing Time:", str(datetime.timedelta(seconds=processing))))
            click.echo("{:40} {:10}".format("Actual Scan Duration: ", str(datetime.timedelta(seconds=total_calculated_scan_duration))))
            click.echo("{:40} {:10} {}".format("Longest Asset Duration: ", str(datetime.timedelta(seconds=longest_scanned[1])), longest_scanned[0]))
            click.echo("{:40} {:10} {}".format("Shortest Asset Duration: ", str(datetime.timedelta(seconds=shortest_scanned[1])), shortest_scanned[0]))
            click.echo("{:40} {}\n".format("Average scan duration per Asset:", str(datetime.timedelta(seconds=asset_average_scantime))))

            click.echo("\nFirst asset scanned started at: {}".format(time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(min(start_scan_timestamp_list)))))
            click.echo("Last asset finished scanning at: {}\n\n".format(time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(max(timestamp_plus_duration_list)))))

        except ZeroDivisionError:
            click.echo("\nScan history had No data.\n")


def get_last_history_id(scanid):
    data = request_data("GET", "/scans/{}/history".format(scanid))
    for scans in data['history']:
        if not scans['is_archived']:
            if scans['status'] == 'completed':
                return scans['id']
            else:
                click.echo("\nNo completed Scan to evaluate\n")
                exit()


def download_csv_by_plugin_id(scan_id, hist_id):

    filename = f'{scan_id}-report.csv'
    click.echo("\nDownloading scan {} details into a csv called {} for parsing and manual auditing\n".format(scan_id, filename))
    tio = tenb_connection()

    # Stream the report to disk
    with open(filename, 'wb') as fobj:
        tio.scans.export(scan_id, ('plugin.id', 'eq', '19506'),
                         format='csv', fobj=fobj, history_id=hist_id)
    parse_19506_from_file(filename, scan_id, hist_id)


def evaluate_a_scan(scanid, histid):

    if scanid:
        if histid:
            download_csv_by_plugin_id(scanid, histid)
        else:
            historyid = get_last_history_id(scanid)
            download_csv_by_plugin_id(scanid, historyid)
    else:
        # Pull all 19506 Plugins from the DB
        plugin_data = db_query("select asset_uuid, output from vulns where plugin_id='19506';")

        # Set some dicts for organizing Data
        scan_policy_dict = {}
        scanner_dict = {}
        scan_name_dict = {}

        # Open a CSV for export
        with open('evaluate.csv', mode='w') as csv_file:
            agent_writer = csv.writer(csv_file, delimiter=',', quotechar='"')

            header_list = ["asset uuid", "Scan Name", "Scan Policy", "Scanner IP", "Scan Time", "Max Checks", "Max Hosts", "Minutes", "RTT", "Hop Count"]

            # Metics Header
            #header_list = ["Policy/Scanner/Scan", "Average per Asset in mins"]

            # Write the header to the csv
            agent_writer.writerow(header_list)

            # This function is used to parse the data
            # Getting the length of the Category and using it to get the average
            def average_by_policy(name, scan_info):
                # Print the Category to the Screen ( Scanner, Policy, Scan Name)
                click.echo("\n{:100s} {:25s} {:10}".format(name, "AVG Minutes Per/Asset", "Total Assets"))
                click.echo("-" * 20)

                # Write the category to the csv for delination
                #agent_writer.writerow([name])

                # Cycle through each category
                for scan in scan_info.items():

                    # data is in a list [asset_uuid, mins] We need the length of the total mins found
                    length = len(scan[1])

                    # Reset the total per Category Item - Specific Scan ID, Scanner, Policy ID
                    total = 0

                    # Cycle through each asset record
                    for assets in scan[1].values():
                        # Gather a total
                        total = assets + total

                    # After calculating the total, lets get an average
                    average = total/length

                    # Print results to the screen
                    click.echo("\n{:100s} {:25d} {:10d}".format(scan[0], int(average), length))

                    update = [scan[0], average]

                    # Send the results as a list to the CSV
                    #agent_writer.writerow(update)
                click.echo("-" * 150)

            # Loop through each plugin 19506 and Parse data from it
            for vulns in plugin_data:

                # Output is the second item in the tuple from the DB
                plugin_output = vulns[1]

                # split the output by return
                parsed_output = plugin_output.split("\n")

                # grab the length so we can grab the seconds
                plugin_length = len(parsed_output)

                # grab the scan duration- second to the last variable
                duration = parsed_output[plugin_length - 2]

                # Split at the colon to grab the numerical value
                seconds = duration.split(" : ")

                # split to remove "secs"
                number = seconds[1].split(" ")

                # grab the number for our minute calculation
                final_number = number[0]

                if final_number != 'unknown':
                    try:
                        # convert seconds into minutes
                        minutes = int(final_number) / 60

                        # Grab data pair and split it at the colon and grab the values
                        scan_name = parsed_output[9].split(" : ")[1]
                        scan_policy = parsed_output[10].split(" : ")[1]
                        scanner_ip = parsed_output[11].split(" : ")[1]
                        scan_time = parsed_output[plugin_length - 3].split(" : ")[1]
                        max_hosts = parsed_output[plugin_length- 8].split(" : ")[1]
                        max_checks = parsed_output[plugin_length - 7].split(" : ")[1]

                        if "no" not in parsed_output[14].split(" : ")[1]:
                            rtt = parsed_output[14].split(" : ")[1]
                        else:
                            rtt = parsed_output[12].split(" : ")[1]

                        try:
                            # Grab the last line in the Trace route Plugin output
                            # Split on the space and grab the numerical value.
                            hopcount = grab_hop_count(vulns[0])[0][-1].split(" ")[-1]
                        except IndexError:
                            hopcount = "Unknown"

                        parsed_data_organized = [vulns[0], scan_name, scan_policy, scanner_ip, scan_time, max_checks, max_hosts, minutes, rtt, hopcount]

                        agent_writer.writerow(parsed_data_organized)

                        # Organize Data by Scan Policy
                        # If the category is not in the new dict, add it; else update it.
                        if scan_policy not in scan_policy_dict:
                            scan_policy_dict[scan_policy] = {vulns[0]: minutes}
                        else:
                            scan_policy_dict[scan_policy].update({vulns[0]: minutes})

                        if scanner_ip not in scanner_dict:
                            scanner_dict[scanner_ip] = {vulns[0]: minutes}
                        else:
                            scanner_dict[scanner_ip].update({vulns[0]: minutes})

                        if scan_name not in scan_name_dict:
                            scan_name_dict[scan_name] = {vulns[0]: minutes}
                        else:
                            scan_name_dict[scan_name].update({vulns[0]: minutes})
                    except IndexError:
                        # This error occurs when an old scanner is used.
                        # the 19506 plugin filled with an error indicating the need for an upgrade
                        pass

            average_by_policy("Policies", scan_policy_dict)

            average_by_policy("Scanners", scanner_dict)

            average_by_policy("Scan Name", scan_name_dict)




import csv
import click
from sqlite3 import Error
from .database import new_db_connection


def compliance_export_csv(name, uuid, file_name):

    database = r"navi.db"
    conn = new_db_connection(database)
    with conn:
        cur = conn.cursor()
        # Since the Compliance export doesn't have the FQDN of the asset, pull it from the asset table
        try:
            if name and uuid:
                cur.execute("SELECT assets.fqdn, compliance.* FROM compliance LEFT OUTER JOIN assets ON "
                            "assets.uuid = compliance.asset_uuid "
                            "where compliance.audit_file='{}' and compliance.asset_uuid='{}';".format(name, uuid))
            elif name:
                cur.execute("SELECT assets.fqdn, compliance.* FROM compliance LEFT OUTER JOIN assets ON "
                            "assets.uuid = compliance.asset_uuid "
                            "where audit_file='{}';".format(name))

                file_name = "{}.csv".format(name)
            elif uuid:
                cur.execute("SELECT assets.fqdn, compliance.* FROM compliance LEFT OUTER JOIN assets ON "
                            "assets.uuid = compliance.asset_uuid "
                            "where asset_uuid='{}';".format(uuid))

                file_name = "{}.csv".format(uuid)
            else:
                cur.execute("SELECT assets.fqdn, compliance.* FROM compliance LEFT OUTER JOIN assets ON "
                            "assets.uuid = compliance.asset_uuid;")
        except Error:
            print("\n No data! \n Please run 'navi update compliance' first\n")

        data = cur.fetchall()

        click.echo("I'm exporting {} with your requested data\n".format(file_name))

        header_list = ["FQDN", "asset_uuid", "actual_value", "audit_file", "check_id", "check_info", "check_name",
                       "expected_value", "first_seen", "last_seen", "plugin_id", "reference", "see_also", "solution",
                       "status"]

        # Crete a csv file object
        with open(file_name, mode='w', encoding='utf-8', newline="") as csv_file:
            compliance_writer = csv.writer(csv_file, delimiter=',', quotechar='"')

            compliance_writer.writerow(header_list)
            # Loop through each asset
            for assets in data:
                compliance_writer.writerow(assets)



import time
import threading
import click
from queue import Queue
from sqlite3 import Error
from .api_wrapper import request_data
from .database import new_db_connection, insert_assets, insert_tags, drop_tables, get_last_update_id, insert_update_info
from .dbconfig import create_assets_table, create_tag_table

lock = threading.Lock()

q = Queue()

tag_id = 0


def worker():
    # The worker thread pulls an item from the queue and processes it
    while True:
        item = q.get()
        parse_data(request_data('GET', item))
        q.task_done()


def parse_data(chunk_data):
    database = r"navi.db"
    asset_conn = new_db_connection(database)
    asset_conn.execute('pragma journal_mode=wal;')

    with asset_conn:
        for assets in chunk_data:
            # create a blank list to append asset details
            csv_list = []

            # Capture the first IP
            try:
                ip = assets['ipv4s'][0]
                csv_list.append(ip)
            except IndexError:
                csv_list.append(" ")

            try:
                csv_list.append(assets['hostnames'][0])
            except IndexError:
                csv_list.append(" ")

            try:
                csv_list.append(assets['fqdns'][0])
            except IndexError:
                csv_list.append(" ")

            try:
                asset_id = assets['id']
                csv_list.append(asset_id)
            except KeyError:
                csv_list.append(" ")
            try:

                csv_list.append(assets['first_seen'])
            except KeyError:
                csv_list.append(" ")
            try:

                csv_list.append(assets['last_seen'])
            except KeyError:
                csv_list.append(" ")
            try:
                csv_list.append(assets['operating_systems'][0])
            except IndexError:
                csv_list.append(" ")

            try:
                csv_list.append(assets['mac_addresses'][0])
            except IndexError:
                csv_list.append(" ")

            try:
                csv_list.append(assets['agent_uuid'])
            except KeyError:
                csv_list.append(" ")

            try:
                csv_list.append(assets["last_licensed_scan_date"])
            except KeyError:
                csv_list.append(" ")

            try:
                csv_list.append(assets["network_id"])
            except KeyError:
                csv_list.append(" ")

            try:
                csv_list.append(assets["acr_score"])
            except KeyError:
                csv_list.append(" ")

            try:
                csv_list.append(assets["exposure_score"])
            except KeyError:
                csv_list.append(" ")

            try:
                csv_list.append(assets['aws_ec2_instance_id'])
            except KeyError:
                csv_list.append(" ")

            try:
                csv_list.append(assets['aws_ec2_instance_state'])
            except KeyError:
                csv_list.append(" ")

            try:
                csv_list.append(assets['aws_ec2_name'])
            except KeyError:
                csv_list.append(" ")

            try:
                csv_list.append(assets['aws_ec2_region'])
            except KeyError:
                csv_list.append(" ")

            try:
                csv_list.append(assets['aws_availability_zone'])
            except KeyError:
                csv_list.append(" ")

            try:
                csv_list.append(assets['gcp_instance_id'])
            except KeyError:
                csv_list.append(" ")

            try:
                csv_list.append(assets['gcp_project_id'])
            except KeyError:
                csv_list.append(" ")

            try:
                csv_list.append(assets['gcp_zone'])
            except KeyError:
                csv_list.append(" ")

            try:
                special_url = "https://cloud.tenable.com/tio/app.html#/vulnerability-management/dashboard/assets/asset-details/{}/vulns".format(asset_id)
                csv_list.append(special_url)
            except:
                special_url = None
                csv_list.append(special_url)

            # Collect and save Tag Data
            try:
                global tag_id
                tag_ip = assets['ipv4s'][0]
                tag_asset_id = assets['id']
                for t in assets["tags"]:
                    tag_list = []
                    tag_id = tag_id + 1
                    tag_list.append(tag_id)
                    tag_list.append(tag_asset_id)
                    tag_list.append(tag_ip)

                    tag_key = t['key']
                    tag_list.append(tag_key)

                    tag_uuid = t['uuid']
                    tag_list.append(tag_uuid)

                    tag_value = t['value']
                    tag_list.append(tag_value)

                    tag_added_date = t['added_at']
                    tag_list.append(tag_added_date)

                    try:
                        insert_tags(asset_conn, tag_list)
                    except Error as e:
                        click.echo(e)
            except IndexError:
                pass
            try:
                insert_assets(asset_conn, csv_list)
            except Error as e:
                click.echo(e)
    asset_conn.close()


def asset_export(days, ex_uuid, threads, category, value):
    start = time.time()

    # Crete a new connection to our database
    database = r"navi.db"
    drop_conn = new_db_connection(database)
    drop_tables(drop_conn, 'assets')
    drop_tables(drop_conn, 'tags')

    # Set URLS for threading
    urls = []

    # Set the payload to the maximum number of assets to be pulled at once
    day = 86400
    new_limit = day * int(days)
    day_limit = time.time() - new_limit
    if category is None:
        pay_load = {"chunk_size": 1000, "filters": {"updated_at": int(day_limit)}}
    else:
        if value is None:
            pay_load = {"chunk_size": 1000, "filters": {"updated_at": int(day_limit)}}
        else:
            pay_load = {"chunk_size": 1000, "filters": {"updated_at": int(day_limit), "tag.{}".format(category): value}}
    try:

        if ex_uuid == '0':
            # request an export of the data
            export = request_data('POST', '/assets/export', payload=pay_load)

            # grab the export UUID
            ex_uuid = export['export_uuid']
            click.echo('\nRequesting Asset Export with ID : {}'.format(ex_uuid))

            # set a variable to True for our While loop
            not_ready = True
        else:
            click.echo("\nUsing your Export UUID\n")
            not_ready = False

        # now check the status
        status = request_data('GET', '/assets/export/' + ex_uuid + '/status')

        # status = get_data('/vulns/export/89ac18d9-d6bc-4cef-9615-2d138f1ffsdf/status')
        click.echo("Status : {}".format(str(status["status"])))

        # loop to check status until finished
        while not_ready is True:
            # Pull the status, then pause 5 seconds and ask again.
            if status['status'] == 'PROCESSING' or 'QUEUED':
                time.sleep(2.5)
                status = request_data('GET', '/assets/export/' + ex_uuid + '/status')
                # click.echo("Status : " + str(status["status"]))

            # Exit Loop once confirmed finished
            if status['status'] == 'FINISHED':
                ptime = time.time()
                click.echo("\nProcessing Time took : {}".format(str(ptime - start)))
                not_ready = False

            # Tell the user an error occured
            if status['status'] == 'ERROR':
                click.echo("Error occurred")

        create_assets_table()
        create_tag_table()

        # grab all of the chunks and craft the URLS for threading

        for y in status['chunks_available']:
            urls.append('/assets/export/' + ex_uuid + '/chunks/' + str(y))

        for i in range(threads):
            t = threading.Thread(target=worker)
            t.daemon = True  # thread dies when main thread (only non-daemon thread) exits.
            t.start()

        # stuff work items on the queue (in this case, just a number).
        # start = time.perf_counter()
        for item in range(len(urls)):
            q.put(urls[item])

        q.join()
        end = time.time()
        click.echo("Asset Download took: {}\n".format(str(end - start)))

        # Now that the download has completed we need to record it
        update_id = get_last_update_id()
        diff_dict = [update_id, str(start), str(days), "Asset Update", str(ex_uuid)] # need to ignore if the Ex_uuid exsists in the db.
        database_2 = r"navi.db"
        conn = new_db_connection(database_2)
        with conn:
            insert_update_info(conn, diff_dict)

    except IndexError:
        click.echo("Well this is a bummer; you don't have permissions to download Asset data :( ")
    except TypeError:
        click.echo("You may not be authorized or your keys are invalid")




import sqlite3
from sqlite3 import Error
import click


def new_db_connection(db_file):
    # create a connection to our database
    conn = None
    try:
        # A database file will be created if one doesn't exist
        conn = sqlite3.connect(db_file, timeout=10.0)
    except Error as E:
        click.echo(E)
    return conn


def create_table(conn, table_information):
    try:
        c = conn.cursor()
        c.execute('pragma journal_mode=wal;')
        c.execute(table_information)
    except Error as e:
        click.echo(e)


def db_query(statement):
    # start = time.time()
    database = r"navi.db"
    query_conn = new_db_connection(database)
    with query_conn:
        cur = query_conn.cursor()
        cur.execute('pragma journal_mode=wal;')
        cur.execute('pragma cache_size=-10000;')
        cur.execute('PRAGMA synchronous = OFF')
        cur.execute('pragma threads=4')
        cur.execute(statement)

        data = cur.fetchall()
        # end = time.time()
        # total = end - start
    query_conn.close()
    # click.echo("Sql Query took: {} seconds".format(total))
    return data


def get_last_update_id():
    database = r"navi.db"
    conn = new_db_connection(database)
    try:
        with conn:
            cur = conn.cursor()
            cur.execute("SELECT update_id from diff;")
            data = cur.fetchall()

            new_id = len(data) + 1
    except Error:
        new_id = 1
    return new_id


def insert_update_info(conn, diff):
    sql = '''INSERT or IGNORE into diff(update_id, timestamp, days, update_type, exid) VALUES(?,?,?,?,?)'''
    cur = conn.cursor()
    cur.execute('pragma journal_mode=wal;')
    cur.execute(sql, diff)


def insert_compliance(conn, compliance):
    sql = '''INSERT or IGNORE into compliance(asset_uuid, actual_value, audit_file, check_id, check_info, check_name, 
                    expected_value, first_seen, last_seen, plugin_id, reference, see_also, solution, status) VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?)'''
    cur = conn.cursor()
    cur.execute('pragma journal_mod=wal;')
    cur.execute(sql, compliance)


def insert_assets(conn, assets):
    sql = '''INSERT or IGNORE into assets(
                                          ip_address, 
                                          hostname, 
                                          fqdn, 
                                          uuid, 
                                          first_found, 
                                          last_found, 
                                          operating_system,
                                          mac_address, 
                                          agent_uuid, 
                                          last_licensed_scan_date, 
                                          network, 
                                          acr, 
                                          aes, 
                                          aws_id,
                                          aws_ec2_instance_state,
                                          aws_ec2_name,
                                          aws_ec2_region,
                                          aws_availability_zone,
                                          gcp_instance_id,
                                          gcp_project_id,
                                          gcp_zone,
                                          url) VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)'''
    cur = conn.cursor()
    cur.execute('pragma journal_mode=wal;')
    cur.execute(sql, assets)


def insert_tags(conn, tags):
    sql = '''INSERT or IGNORE into tags(tag_id, asset_uuid, asset_ip, tag_key, tag_uuid, tag_value, tag_added_date) VALUES(?,?,?,?,?,?,?)'''
    cur = conn.cursor()
    cur.execute('pragma journal_mode=wal;')
    cur.execute(sql, tags)


def drop_tables(conn, table):
    try:
        drop_table = '''DROP TABLE {}'''.format(table)
        cur = conn.cursor()
        cur.execute('pragma journal_mode=wal;')
        cur.execute(drop_table)
    except Error:
        pass


def insert_vulns(conn, vulns):
    sql = '''INSERT or IGNORE into vulns(
                            asset_ip, 
                            asset_uuid, 
                            asset_hostname, 
                            first_found, 
                            last_found, 
                            output, 
                            plugin_id, 
                            plugin_name, 
                            plugin_family, 
                            port, 
                            protocol, 
                            severity, 
                            scan_completed, 
                            scan_started, 
                            scan_uuid, 
                            schedule_id, 
                            state,
                            cves,
                            score,
                            exploit,
                            xrefs,
                            synopsis, 
                            see_also,
                            solution,
                            version, 
                            description, 
                            cvss3_base_score,
                            cvss3_temporal_score,
                            cvss_base_score,
                            cvss_temporal_score,
                            OSes,
                            url
    ) VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)'''

    cur = conn.cursor()
    cur.execute('pragma journal_mode=wal;')
    cur.execute(sql, vulns)


def insert_apps(conn, apps):
    sql = '''INSERT or IGNORE into apps(
             name,
             uuid, 
             target, 
             scan_completed_time,
             pages_audited,
             pages_crawled,
             requests_made, 
             critical_count,
             high_count,
             medium_count,
             low_count, 
             info_count,
             owasp,
             tech_list,
             config_id)
     VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)'''
    cur = conn.cursor()
    cur.execute('pragma journal_mode=wal;')
    cur.execute(sql, apps)


def insert_fixed(conn, fixed):
    sql = '''INSERT or IGNORE into fixed(
                            asset_uuid, 
                            output, 
                            plugin_id, 
                            plugin_name, 
                            port,
                            first_found,
                            last_fixed,
                            last_found,
                            severity,
                            delta,
                            pass_fail,
                            state,
                            special_url
    ) VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?)'''

    cur = conn.cursor()
    cur.execute('pragma journal_mode=wal;')
    cur.execute(sql, fixed)



import csv
import time
from .api_wrapper import tenb_connection

tio = tenb_connection()


def agent_group_export(group):
    try:
        for group_info in tio.agent_groups.list():
            group_name = group_info['name']

            if group_name == group:
                group_id = group_info['id']

                with open('agent_group_data.csv', mode='w', encoding='utf-8', newline="") as csv_file:
                    agent_writer = csv.writer(csv_file, delimiter=',', quotechar='"')

                    header_list = ["Agent Name", "IP Address", "Platform", "Last connected", "Last scanned", "Status",
                                   "Plugin Feed ID"]

                    agent_writer.writerow(header_list)

                    agents = tio.agent_groups.details(group_id)

                    for agent in agents['agents']:
                        try:
                            name = agent['name']
                            ip = agent['ip']
                            platform = agent['platform']
                        except KeyError:
                            name = "Unknown"
                            ip = "Unknown"
                            platform = "Unknown"

                        try:
                            plugin_feed = agent['plugin_feed_id']
                        except KeyError:
                            plugin_feed = "Plugins Not updated"

                        try:
                            last_connect = agent['last_connect']
                            connect_time = time.strftime("%a, %d %b %Y %H:%M:%S", time.localtime(last_connect))
                        except KeyError:
                            connect_time = "Unknown"

                        try:
                            last_scanned = agent['last_scanned']
                            scanned_time = time.strftime("%a, %d %b %Y %H:%M:%S", time.localtime(last_scanned))
                        except KeyError:
                            scanned_time = "Not Yet Scanned"

                        status = agent['status']
                        agent_writer.writerow([name, ip, platform, connect_time, scanned_time, status, plugin_feed])
    except AttributeError:
        print("\nCheck your permissions or your API keys\n")
    return



import csv
import time
from .api_wrapper import tenb_connection

tio = tenb_connection()


def agent_export():

    with open('agent_data.csv', mode='w', encoding='utf-8', newline="") as csv_file:
        agent_writer = csv.writer(csv_file, delimiter=',', quotechar='"')

        header_list = ["Agent Name", "IP Address", "Platform", "Last connected", "Last scanned", "Status", "Plugin Feed ID"]
        agent_writer.writerow(header_list)

        try:
            for agent in tio.agents.list(limit=5000):

                try:
                    name = agent['name']
                    ip = agent['ip']
                    platform = agent['platform']
                except KeyError:
                    name = "Unknown"
                    ip = "Unknown"
                    platform = "Unknown"

                try:
                    plugin_feed = agent['plugin_feed_id']
                except KeyError:
                    plugin_feed = "Plugins Not updated"

                try:
                    last_connect = agent['last_connect']
                    connect_time = time.strftime("%a, %d %b %Y %H:%M:%S", time.localtime(last_connect))
                except KeyError:
                    connect_time = "Unknown"

                try:
                    last_scanned = agent['last_scanned']
                    scanned_time = time.strftime("%a, %d %b %Y %H:%M:%S", time.localtime(last_scanned))
                except KeyError:
                    scanned_time = "Not Yet Scanned"

                status = agent['status']
                agent_writer.writerow([name, ip, platform, connect_time, scanned_time, status, plugin_feed])
        except AttributeError:
            print("\nCheck your permissions or your API keys\n")
    return



import click
from os import system as sys
from .database import db_query
import pprint


@click.group(help="Run Navi Tag Rules")
def rules():
    pass


@rules.command()
def run():
    try:
        rule_data = db_query("select * from rules;")

        pprint.pprint(rule_data)

        for rule in rule_data:
            if rule[3] == 'plugin_id':

                print("navi tag --c \"{}\" --v \"{}\" --plugin {}".format(rule[1], rule[2], rule[4]))
                sys("navi tag --c \"{}\" --v \"{}\" --plugin {}".format(rule[1], rule[2], rule[4]))
            elif rule[3] == 'plugin_name':
                print("navi tag --c \"{}\" --v \"{}\" --name \"{}\"".format(rule[1], rule[2], rule[4]))
                sys("navi tag --c \"{}\" --v \"{}\" --name \"{}\"".format(rule[1], rule[2], rule[4]))
            elif rule[3] == 'plugin_output':
                print("navi tag --c \"{}\" --v \"{}\" --plugin {} --output {}".format(rule[1], rule[2], rule[5], rule[4]))
                sys("navi tag --c \"{}\" --v \"{}\" --plugin {} --output {}".format(rule[1], rule[2], rule[5], rule[4]))
            elif rule[3] == 'cve':
                print("navi tag --c \"{}\" --v \"{}\" --cve {}".format(rule[1], rule[2], rule[4]))
                sys("navi tag --c \"{}\" --v \"{}\" --cve {}".format(rule[1], rule[2], rule[4]))
            elif rule[3] == 'xref':
                print("navi tag --c \"{}\" --v \"{}\" --xref {}".format(rule[1], rule[2], rule[4]))
                sys("navi tag --c \"{}\" --v \"{}\" --xref {}".format(rule[1], rule[2], rule[4]))
    except:
        click.echo("\nYou don't have any Rules")



import click
from .database import db_query
from .api_wrapper import request_data, tenb_connection
from .usergroup import get_group_id
from .user import get_user_id
import pprint


tio = tenb_connection()


def create_permission(name, tag_name, uuid, perm_string, perm_type, subject_type):
    payload = {
        "actions": ["{}".format(perm_string)],
        "objects": [
            {
                "name": name,
                "type": perm_type,
                "uuid": str(uuid)
            }
        ],
        "subjects": [{"type": subject_type}],
        "name": "{} : [{}]".format(tag_name, perm_string)
    }
    response = request_data("POST", "/api/v3/access-control/permissions", payload=payload)
    return response


def create_granular_permission(tag_name, uuid, perm_list, perm_type, subject_type, subject_name, subject_uuid):
    payload = {
        "actions": perm_list,
        "objects": [
            {
                "name": tag_name,
                "type": perm_type,
                "uuid": str(uuid)
            }
        ],
        "subjects": [{"name": subject_name, "type": subject_type, "uuid": subject_uuid}],
        "name": "{} : {}".format(tag_name, perm_list)
    }
    response = request_data("POST", "/api/v3/access-control/permissions", payload=payload)
    return response


def grab_can_use_tags():
    # This script supports the Migrate command which adds CanUse to ALLUsers
    data = request_data("GET", "/api/v3/access-control/permissions")

    list_of_tag_uuids = []
    # Filter on just permissions
    for perms in data['permissions']:
        # Need to search for CanUse
        if 'CanUse' in str(perms['actions']):

            # Extract all Tag UUIDs
            for tag in perms['objects']:
                try:
                    if tag['type'] == 'Tag':
                        list_of_tag_uuids.append(tag['uuid'])
                except KeyError:
                    pass
    return list_of_tag_uuids


@click.group(help="Change Access Control Permissions")
def access():
    pass


@access.command(help="Change Access Control Permissions using a Tag")
@click.option('--c', default='', required=True, help="Tag Category name to use")
@click.option('--v', default='', required=True, help="Tag Value to use")
@click.option('--user', default='', help="The User you want to assign to the Permission")
@click.option('--usergroup', default='', help="The User Group you want to assign to the Permission")
@click.option('--perm', multiple=True, type=click.Choice(['CanScan', 'CanView', 'CanEdit', 'CanUse'], case_sensitive=True),
              required=True)
def create(c, v, user, usergroup, perm):
    # Create the naming format for the tag permission
    perm_name = "{},{}".format(c, v)
    try:

        tag_uuid = 0
        # Grab the Tag UUID using the value and category given
        for tag in tio.tags.list():

            tag_value = tag['value']
            tag_category = tag['category_name']
            if c == tag_category and v == tag_value:
                tag_uuid = tag['uuid']

        # Add permission by User
        if user:
            user_id, uuid = get_user_id(user)
            resp = create_granular_permission(tag_name=perm_name, uuid=tag_uuid,
                                              perm_list=perm, perm_type="Tag", subject_type="User",
                                              subject_name=user, subject_uuid=uuid)
            pprint.pprint(resp)

        # Add permission by UserGroup
        elif usergroup:
            group_id, uuid = get_group_id(usergroup)
            resp = create_granular_permission(tag_name=perm_name, uuid=tag_uuid,
                                              perm_list=perm, perm_type="Tag", subject_type="UserGroup",
                                              subject_name=usergroup, subject_uuid=uuid)
            pprint.pprint(resp)
        else:
            # If no user or Usergroup Assign it to All Users
            permission_response = create_permission(name=perm_name, tag_name=perm_name, uuid=tag_uuid,
                                                    perm_string=perm, perm_type="Tag", subject_type="AllUsers")

            pprint.pprint(permission_response)

    except IndexError:
        click.echo("\nYour Tag might be incorrect. Or you may need to update assets in navi.  "
                   "No tag UUID was returned\n")
        exit()


@access.command(help='Find all Tags without Permissions and apply "CanUse" permissions to AllUsers')
def migrate():
    click.confirm("\n This script finds all tags without any permissions and creates a CANUSE permission. "
                  "Do you want to continue?\n")
    unique_ids = []
    new_list_uuids = []

    # Grab all of the Tags with Can Use
    canuse_tags = grab_can_use_tags()

    # Grab all Tags and their names from the navi db
    tag_in_db = db_query("select tag_uuid, tag_key, tag_value from tags;")

    for tag in tag_in_db:
        # reduce the amount of tags uuids, since the navi db has a record for each asset:tag_uuid
        if tag[0] not in unique_ids:
            unique_ids.append(tag[0])
            # create a dict to hold the tag name and the tag uuid
            new_list_uuids.append({"{},{}".format(tag[1], tag[2]): tag[0]})

    # Loop through the list of uuids without permissions
    click.echo("\nCanUse permissions will be created for each of these tags\n")
    for uuid in new_list_uuids:
        for key in uuid:
            if uuid[key] not in canuse_tags:
                # Add permissions for each tag.

                print(str(key), str(uuid[key]))
                create_permission(name=key, tag_name=key, uuid=uuid[key], perm_string="CanUse", perm_type="Tag",
                                  subject_type="AllUsers")

    print()
